using System;
using System.Collections.Generic;
using System.Text;
using System.Net.Sockets;

namespace SMPP
{
    public class AsyncSocketClient
    {
        #region delegates
        public delegate void MessageHandler(AsyncSocketClient asyncSocketClient);
        public delegate void SocketClosingHandler(AsyncSocketClient asyncSocketClient);
        public delegate void ErrorHandler(AsyncSocketClient asyncSocketClient, Exception exception);
        #endregion

        private const int BUFFER_SIZE = 1048576;
        
        #region fields

        private NetworkStream networkStream;
        private TcpClient tcpClient;
        private AsyncCallback callBackReadMethod;
        private AsyncCallback callBackWriteMethod;
        private MessageHandler messageHandler;
        private SocketClosingHandler socketCloseHandler;
        private ErrorHandler errorHandler;
        private bool isDisposed;
        private string serverIPAddress;
        private Int16 serverPort;
        private object stateObject;
        private byte[] buffer;

        #endregion

        #region properties

        public string ServerIPAddress
        {
            get
            {
                return this.serverIPAddress;
            }
        }

        public Int16 ServerPort
        {
            get
            {
                return this.serverPort;
            }
        }

        public object StateObject
        {
            get
            {
                return this.stateObject;
            }
            set
            {
                this.stateObject = value;
            }
        }

        public byte[] Buffer
        {
            get
            {
                return this.buffer;
            }
        }

        #endregion 

        #region constructor
        public AsyncSocketClient(Int32 bufferSize, object stateObject,
                                 MessageHandler msgHandler, SocketClosingHandler closingHandler,
                                 ErrorHandler errHandler)
        {
            this.buffer = new byte[BUFFER_SIZE];
            this.stateObject = stateObject;

            this.messageHandler = msgHandler;
            this.socketCloseHandler = closingHandler;
            this.errorHandler = errHandler;

            this.callBackReadMethod = new AsyncCallback(ReceiveComplete);
            this.callBackWriteMethod = new AsyncCallback(SendComplete);

            this.isDisposed = false;
        }

        ~AsyncSocketClient()
        {
            if (!this.isDisposed)
            {
                Dispose();
            }
        }
        #endregion

        public void Dispose()
        {
            try
            {
                this.isDisposed = true;
                Disconnect();
            }
            catch
            { }
        }

        public void Connect(String IPAddress, Int16 port)
        {
            try
            {
                if (this.networkStream == null)
                {
                    this.serverIPAddress = IPAddress;
                    this.serverPort = port;

                    this.tcpClient = new TcpClient(this.serverIPAddress, this.serverPort);
                    this.networkStream = this.tcpClient.GetStream();

                    this.tcpClient.ReceiveBufferSize = BUFFER_SIZE;
                    this.tcpClient.SendBufferSize = BUFFER_SIZE;
                    this.tcpClient.NoDelay = true;

                    this.tcpClient.LingerState = new LingerOption(false, 0);

                    Receive();
                }
            }
            catch (SocketException exc)
            {
                throw new Exception(exc.Message, exc.InnerException);
            }
        }

        public void Disconnect()
        {
            if (this.networkStream != null)
            {
                this.networkStream.Close();
            }
            if (this.tcpClient != null)
            {
                this.tcpClient.Close();
            }

            this.networkStream = null;
            this.tcpClient = null;
        }

        public void Send(byte[] buffer)
        {
            try
            {
                if (this.networkStream != null && this.networkStream.CanWrite)
                {
                    this.networkStream.BeginWrite(buffer, 0, buffer.Length, this.callBackWriteMethod, null);
                }
                else
                {
                    throw new Exception("Socket is closed, cannot Send().");
                }
            }
            catch
            {
                throw;
            }
        }

        public void Receive()
        {
            try
            {
                if (this.networkStream != null && this.networkStream.CanRead)
                {
                    this.buffer = new byte[BUFFER_SIZE];

                    this.networkStream.BeginRead(this.buffer, 0, this.buffer.Length, this.callBackReadMethod, null);
                }
                else
                {
                    throw new Exception("Socket is closed,cannot Receive().");
                }
            }
            catch
            {
                throw;
            }
        }

        private void SendComplete(IAsyncResult state)
        {
            try
            {
                if (this.networkStream.CanWrite)
                {
                    this.networkStream.EndWrite(state);
                }
            }
            catch
            { }
        }

        private void ReceiveComplete(IAsyncResult state)
        {
            try
            {
                if (this.networkStream.CanRead)
                {
                    int bytesReceived = this.networkStream.EndRead(state);

                    if(bytesReceived >0 )
                    {
                        try
                        {
                            this.messageHandler(this);
                        }
                        finally
                        {
                            Receive();
                        }
                    }
                }
            }
            catch
            {
                try
                {
                    this.socketCloseHandler(this);
                }
                finally
                {
                    Dispose();
                }
            }
        }


   }
}
