using System;
using System.Collections;
using System.Diagnostics;
using SMPP.Utility;
using System.Text;

namespace SMPP.Packet
{
    public abstract class Pdu : IDisposable
    {
        #region Constatnts

        protected const int HEADER_LENGTH = 16;
        protected const int DATE_TIME_LENGTH = 16;

        #endregion constants

        #region Fields

        private static uint staticSequenceNumber =0;
        private uint commandStatus;
        private CommandIdType commandID;
        private TlvTable tlvTable = new TlvTable();
        private uint customSequenceNumber = 0;
        private uint sequenceNumber;
        private uint commandLength;
        private byte[] packetBytes = new byte[0];

        #endregion Fields

        #region Properties
 
        public byte[] PacketBytes
        {
            get
            {
                return(byte[])packetBytes.Clone();
            }
            set
            {
                packetBytes = value;
            }
        }
        public uint CommandLength
        {
            get
            {
                return this.commandLength;
            }
            set
            {
                this.commandLength = value;
            }
        }
        public uint SequenceNumber
        {
            get
            {
                return this.sequenceNumber;
            }
            set
            {
                customSequenceNumber = value;
            }
        }

        public uint CommandStatus
        {
            get
            {
                return this.commandStatus;
            }
            set
            {
                this.commandStatus = value;
            }
        }

        protected CommandIdType CommandID
        {
            get
            {
                return this.commandID;
            }

            set
            {
                this.commandID = value;
            }
        }

        #endregion

        #region Constructor

        protected Pdu()
        {
            InitPdu();
        }

        protected Pdu(byte[] incomingBytes)
        {
            this.packetBytes = incomingBytes;
            this.commandLength = DecodeCommandLength(this.packetBytes);
            this.CommandID = DecodeCommandId(this.packetBytes);
            this.commandStatus = UnsignedNumConverter.SwapByteOrdering(BitConverter.ToUInt32(this.packetBytes,8));
            this.sequenceNumber = UnsignedNumConverter.SwapByteOrdering(BitConverter.ToUInt32(this.packetBytes,12));
            this.packetBytes = TrimResponsePdu(this.packetBytes);

            DecodeSmscResponse();
         }
        

        #endregion 

        #region Overridable methods
        
        protected virtual void InitPdu()
        {
            CommandStatus = 0;
            CommandID = Pdu.CommandIdType.generic_nack;
        }

        public virtual void ToMsbHexEncoding()
        {
            throw new NotImplementedException("GetMsbHexEncoding is not implemented in Pdu.");
        }

        protected virtual void DecodeSmscResponse()
        {
            throw new NotImplementedException("DecodeSmsc is not implemented in Pdu.");
        }

        #endregion overridable methods
           
        protected static ArrayList InsertLengthIntoPdu(ArrayList pdu)
        {
            pdu.TrimToSize();
            uint commandLength = (uint)(4 + pdu.Count);
            uint reqLenH2N = UnsignedNumConverter.SwapByteOrdering(commandLength);
            byte[] reqLenArray = BitConverter.GetBytes(reqLenH2N);
            pdu.InsertRange(0,reqLenArray);
            pdu.TrimToSize();

            return pdu;
     
        }

        protected byte[] EncodePduForTrasmission(ArrayList pdu)
        {
            AddTlvBytes(ref pdu);
            pdu = InsertLengthIntoPdu(pdu);
            byte[] result = new byte[pdu.Count];
            pdu.CopyTo(result);

            return result;
        }
        protected UInt16 GetHostOrderUInt16FromTlv(ushort tag)
        {
            return UnsignedNumConverter.SwapByteOrdering(BitConverter.ToUInt16(GetOptionalParamBytes(tag),0));
        }
        protected UInt32 GetHostOrderUInt32FromTlv(ushort tag)
        {
            return UnsignedNumConverter.SwapByteOrdering(BitConverter.ToUInt32(GetOptionalParamBytes(tag),0));
        }
        
        protected void SetHostOrderValueIntoTlv(UInt16 tag,UInt16 val)
        {
            SetOptionalParamBytes(tag,BitConverter.GetBytes(UnsignedNumConverter.SwapByteOrdering(val)));
        }
        protected void SetHostOrderValueIntoTlv(UInt16 tag,UInt32 val)
        {
            SetOptionalParamBytes(tag,BitConverter.GetBytes(UnsignedNumConverter.SwapByteOrdering(val)));
        }

        protected byte[] BytesAfterHeader
        {
            get
            {
                long length = PacketBytes.Length - HEADER_LENGTH;
                byte[] remainder = new byte[length];
                Array.Copy(this.packetBytes,HEADER_LENGTH,remainder,0,length);
                return remainder;
            }
        }

        protected ArrayList GetPduHeader()
        {
            ArrayList pdu = new ArrayList();
            pdu.AddRange(BitConverter.GetBytes(UnsignedNumConverter.SwapByteOrdering((uint)this.CommandID)));
            pdu.AddRange(BitConverter.GetBytes(UnsignedNumConverter.SwapByteOrdering(this.commandStatus)));
            GenerateSequenceNumber();
            pdu.AddRange(BitConverter.GetBytes(UnsignedNumConverter.SwapByteOrdering(SequenceNumber)));
            pdu.TrimToSize();
            return pdu;
        }

        private void GenerateSequenceNumber()
        {
            if(this.customSequenceNumber == 0)
            {
                staticSequenceNumber++;
                if(staticSequenceNumber >= UInt32.MaxValue)
                {
                    staticSequenceNumber = 1;
                }
                this.sequenceNumber = staticSequenceNumber;
            }
            else
            {
                this.sequenceNumber = customSequenceNumber;
            }
        }

        #region TLV table methods
        
        public string GetOptionalParamString(UInt16 tag)
        {
            return this.tlvTable.GetOptionalParamString(UnsignedNumConverter.SwapByteOrdering(tag));
        }

        public byte[] GetOptionalParamBytes(UInt16 tag)
        {
            return this.tlvTable.GetOptionalParamBytes(UnsignedNumConverter.SwapByteOrdering(tag));
        }

        public void SetOptionalParamString(UInt16 tag, string val)
        {
            tlvTable.SetOptionalParamString(UnsignedNumConverter.SwapByteOrdering(tag),val);
        }

        public void SetOptionalParamBytes(UInt16 tag,byte[] val)
        {
            tlvTable.SetOptionalParamBytes(UnsignedNumConverter.SwapByteOrdering(tag),val);
        }

        public void UpdateOptionalParamBytes(UInt16 tag, byte[] val)
        {
            tlvTable.UpdateOptionalParamBytes(UnsignedNumConverter.SwapByteOrdering(tag),val);
        }

        public void UpdateOptionalParamString(UInt16 tag, string val)
        {
            tlvTable.UpdateOptionalParamString(UnsignedNumConverter.SwapByteOrdering(tag),val);
        }

        protected void TranslateTlvDataIntoTable(byte[] tlvBytes)
        {
            tlvTable.TranslateTlvDataIntoTable(tlvBytes);
        }

        protected void TranslateTlvDataIntoTable(byte[] tlvBytes,Int32 index)
        {
            tlvTable.TranslateTlvDataIntoTable(tlvBytes,index);
        }

        protected void AddTlvBytes(ref ArrayList pdu)
        {
            ArrayList tlvs = tlvTable.GenerateByteEncodedTlv();
            foreach (byte[] tlv in tlvs)
            {
                pdu.AddRange(tlv);
            }
        }

       
        #endregion
        
        #region Enumeraion
        public enum BearerType : byte
        {
           Unknown = 0x00,
           SMS = 0x01,
           CircuitSwitchedData = 0x02,
           PacketData = 0x03,
           USSD = 0x04,
           CDPD = 0x05,
           DataTAC = 0x06,
           FLEX_ReFLEX = 0x07,
           CellBroadcast = 0x08
        }
        public enum NetworkType : byte
        {
            Unknown = 0x00,
            GSM = 0x01,
            ANSI_136_TDMA = 0x02,
            IS_95_CDMA = 0x03,
            PDC = 0x04,
            PHS = 0x05,
            iDEN = 0x06,
            AMPS = 0x07,
            PagingNetwork = 0x08
        }

        public enum MessageStateType : byte
        {
            Enroute = 1,
            Delivered = 2,
            Expired = 3,
            Deleted = 4,
            Undelivered = 5,
            Accepted = 6,
            Unknown = 7,
            Rejected = 8
        }
        
        public enum SmppVersionType : byte 
        {
            Version3_3 = 0x33,
            Version3_4 = 0x34
        }

        public enum AddressSubunitType : byte
        {
            Unknown = 0x00,
            MSDisplay = 0x01,
            MobileEquipment = 0x02,
            SmartCard1 = 0x03,
            ExternalUnit1 = 0x04
        }
        
        public enum DisplayTimeType:byte
        {
            Temporary = 0x00,
            Default = 0x01,
            Invoke = 0x02
        }

        public enum ItsReplyTypeType : byte
        {
            Digit = 0,
            Number = 1,
            TelephoneNum = 2,
            Password = 3,
            CharacterLine = 4,
            Menu = 5,
            Date= 5,
            Time = 7,
            Continue = 8
        }
        
        public enum MsValidityType :byte
        {
            StoreIndefinitely = 0x00,
            PowerDown =0x01,
            SIDBased = 0x02,
            DisplayOnly = 0x03
        }

        public enum OptionalParamCodes : ushort
        {
            dest_addr_subunit = 0x0005,
            dest_network_type = 0x0006,
            dest_bearer_type = 0x0007,
            dest_telematics_id = 0x0008,
            source_addr_subunit = 0x000D,
            source_network_type = 0x000E,
            source_bearer_type = 0x000F,
            source_telematics_id = 0x0010,
            qos_time_to_live = 0x0017,
            payload_type = 0x0019,
            additional_status_info_text = 0x001D,
            receipted_message_id = 0x001E,
            ms_msg_wait_failities = 0x0030,
            privacy_indicator = 0x0201,
            source_subaddress = 0x0202,
            dest_subaddress = 0x0203,
            user_message_reference = 0x0204,
            user_response_code = 0x0205,
            source_port = 0x020A,
            destination_port = 0x020B,
            sar_msg_ref_num = 0x020C,
            language_indicator = 0x020D,
            sar_total_segments = 0x020E,
            sar_segment_seqnum = 0x020F,
            SC_interface_version = 0x0210,
            callback_num_pres_ind = 0x0302,
            callback_num_atag = 0x0303,
            number_of_messages=0x0304,
            callback_num = 0x0381,
            dpf_result = 0x0420,
            set_dpf = 0x0421,
            ms_availability_status = 0x0422,
            network_error_code = 0x0423,
            message_payload = 0x0424,
            delivery_faulire_reason = 0x0425,
            more_messages_to_send = 0x0426,
            message_state = 0x0427,
            ussd_service_op=0x0501,
            display_time= 0x1201,
            sms_signal = 0x1203,
            ms_validity = 0x1204,
            alert_on_message_delivery = 0x130C,
            its_replay_type=0x1380,
            its_session_info=0x1383
        }
        public enum TonType : byte
        {
            Unknown = 0x00,
            International = 0x01,
            National = 0x02,
            NetworkSpecific = 0x03,
            SubsciberNumber = 0x04,
            Alphanumeric = 0x05,
            Abbreviated = 0x06
        }
        public enum NpiType : byte
        {
            Unknown = 0x00,
            ISDN = 0x01,
            Data = 0x03,
            Telex = 0x04,
            LandMobile = 0x06,
            National = 0x08,
            Private = 0x09,
            ERMES = 0x0A,
            Internet = 0x0E
        }

        public enum PriorityType : byte
        {
            Lowest = 0x00,
            Level1 = 0x01,
            Level2 = 0x02,
            Hihest = 0x03
        }

        public enum RegisteredDeliveryType : byte
        {
            None = 0x00,
            OnSuccessOrFailure = 0x01,
            OnFailire = 0x02
        }

        public enum DataCodingType : byte
        {
            SMSCDefault = 0x00,
            IA5_ASCII = 0x01,
            OctetUnspecifiedB = 0x02,
            Latin1 = 0x03,
            OctetUnspecifiedA = 0x04,
            JIS = 0x05,
            Cyrillic = 0x06,
            Latin_Hebrew = 0x07,
            UCS2 = 0x08,
            Pictogram = 0x09,
            MusicCodes = 0x0A,
            ExtendedKajiJIS = 0x0D,
            KS_C = 0x0E
        }

        public enum PrivacyType : byte
        {
            Nonrestricted = 0x00,
            Restricted = 0x01,
            Confidental = 0x03,
            Secret = 0x03
        }

        public enum PayloadTypeType : byte
        {
            WPDMessage = 0x00,
            WCMPMessage = 0x01
        }

        public enum LanguageType : byte
        {
            Unspecified = 0x00,
            English = 0x01,
            French = 0x02,
            Spanish = 0x03,
            German = 0x04,
            Portuguese = 0x05
        }
        public enum DpfResultType : byte
        {
            DPFNotSet = 0,
            DPFSet = 1
        }

        public enum CommandIdType: uint
        {
           generic_nack = 0x80000000,
           bind_receiver = 0x00000001,
           bind_receiver_resp = 0x80000001,
           bind_trasmitter = 0x00000002,
           bind_trasmitter_resp = 0x80000002,
           query_sm = 0x00000003,
           query_sm_resp = 0x80000003,
           submit_sm = 0x00000004,
           submit_sm_resp = 0x80000004,
           deliver_sm = 0x00000005,
           deliver_sm_resp = 0x80000005,
           unbind = 0x00000006,
           unbind_resp = 0x80000006,
           replace_sm = 0x00000007,
           replace_sm_resp = 0x80000007,
           cancel_sm = 0x00000008,
           cancel_sm_resp = 0x80000008,
           bind_transceiver = 0x00000009,
           bind_transceiver_resp = 0x80000009,
           outbind = 0x0000000b,
           enquire_link = 0x00000015,
           enquire_link_resp = 0x80000015,
           submit_multi = 0x00000021,
           submit_multi_resp = 0x80000021,
           alert_notification = 0x00000102,
           data_sm = 0x00000103,
           data_sm_resp = 0x80000103,
         }
        #endregion Enumeration

        #region utility methods

        public static byte[] TrimResponsePdu(byte[] response)
        {
            uint commLength = DecodeCommandLength(response);
            if(commLength == response.Length) 
            {
                return response;
            }
            if(commLength >= Int32.MaxValue || commLength > response.Length)
            {
                return new Byte[0];
            }
            byte[] trimmmed = new Byte[commLength];
            for(int i=0;i<trimmmed.Length;i++)
            {
                trimmmed[i] = response[i];
            }
            return trimmmed;
        }

        public static CommandIdType DecodeCommandId(byte[] response)
        {
            uint id = 0;
            try
            {
                id = UnsignedNumConverter.SwapByteOrdering(BitConverter.ToUInt32(response,4));
                return(CommandIdType)id;
            }
            catch
            {
                return CommandIdType.generic_nack;
            }
        }

        public static UInt32 DecodeCommandLength(byte[] response)
        {
            return UnsignedNumConverter.SwapByteOrdering(BitConverter.ToUInt32(response,0));
        }

        #endregion utility methods

        #region IDisposable methods
 
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if(disposing)
            {

            }
        }

        ~Pdu()
        {
            Dispose(false);
        }
        
        #endregion IDisposable methods
        
    }
}
