using System;
using System.Text;
using SMPP.Utility;
using SMPP.Packet;
using System.Collections;

namespace SMPP.Packet.Request
{
	
	public class SmppDeliverSm : Pdu
	{
		#region private fields
		
		private string serviceType = string.Empty;
		private TonType sourceAddressTon = TonType.International;
		private NpiType sourceAddressNpi = NpiType.ISDN;
		private string sourceAddress = string.Empty;
		private TonType destinationAddressTon = TonType.International;
		private NpiType destinationAddressNpi = NpiType.ISDN;
		private string destinationAddress = string.Empty;
		private byte esmClass = 0;
		private SmppVersionType protocolId = SmppVersionType.Version3_4;
		private PriorityType priorityFlag = PriorityType.Lowest;
		private RegisteredDeliveryType registeredDelivery = RegisteredDeliveryType.None;
		private DataCodingType dataCoding = DataCodingType.SMSCDefault;
		private byte smLength = 0;
		private string shortMessage = null;
		
		#endregion private fields
		
		/// <summary>
		/// Used to indicate the SMS Application service associated with the message.
		/// If this is unknown, null is returned.
		/// </summary>
		public string ServiceType
		{
			get
			{
				return this.serviceType;
			}
			
			set
			{
				serviceType = (value == null) ? string.Empty : value;
			}
		}
		
		/// <summary>
		/// Type of Number for source address.
		/// </summary>
		public TonType SourceAddressTon
		{
			get
			{
				return sourceAddressTon;
			}
			
			set
			{
				sourceAddressTon = value;
			}
		}
		
		/// <summary>
		/// Numbering Plan Indicator for source address.
		/// </summary>
		public NpiType SourceAddressNpi
		{
			get
			{
				return sourceAddressNpi;
			}
			
			set
			{
				sourceAddressNpi = value;
			}
		}
		
		/// <summary>
		/// Address of origination entity.
		/// </summary>
		public string SourceAddress
		{
			get
			{
				return sourceAddress;
			}
			
			set
			{
				sourceAddress = (value == null) ? string.Empty : value;
			}
		}
		
		/// <summary>
		/// Type of number of destination entity.
		/// </summary>
		public TonType DestinationAddressTon
		{
			get
			{
				return destinationAddressTon;
			}
			
			set
			{
				destinationAddressTon = value;
			}
		}
		
		/// <summary>
		/// Numbering Plan Indicator of destination entity.
		/// </summary>
		public NpiType DestinationAddressNpi
		{
			get
			{
				return destinationAddressNpi;
			}
			
			set
			{
				destinationAddressNpi = value;
			}
		}
		
		/// <summary>
		/// Destination address of entity.
		/// </summary>
		public string DestinationAddress
		{
			get
			{
				return destinationAddress;
			}
			
			set
			{
				destinationAddress = (value == null) ? string.Empty : value;
			}
		}
		
		/// <summary>
		/// Indicates Message Mode and Message Type.  See the SMSC
		/// version 3.4 specification for details on this.
		/// </summary>
		public byte EsmClass
		{
			get
			{
				return esmClass;
			}
			
			set
			{
				esmClass = value;
			}
		}
		
		/// <summary>
		/// Protocol Identifier; network specific.
		/// </summary>
		public SmppVersionType ProtocolId
		{
			get
			{
				return protocolId;
			}
			
			set
			{
				protocolId = value;
			}
		}
		
		/// <summary>
		/// Designates the priority level of the message.
		/// </summary>
		public PriorityType PriorityFlag
		{
			get
			{
				return priorityFlag;
			}
			
			set
			{
				priorityFlag = value;
			}
		}
		
		/// <summary>
		/// Use this to indicate if you want delivery confirmation.
		/// </summary>
		public RegisteredDeliveryType RegisteredDelivery
		{
			get
			{
				return registeredDelivery;
			}
			
			set
			{
				registeredDelivery = value;
			}
		}
		
		/// <summary>
		/// Indicates the encoding scheme of the short message.
		/// </summary>
		public DataCodingType DataCoding
		{
			get
			{
				return dataCoding;
			}
			
			set
			{
				dataCoding = value;
			}
		}
		
		/// <summary>
		/// Short message length in octets(bytes for x86).
		/// </summary>
		public byte SmLength
		{
			get
			{
				return smLength;
			}
		}
		
		/// <summary>
		/// The short message for this Pdu.  This holds up to 160 characters.
		/// If the message is longer, the MessagePayload property will be used.
		/// If this is the case, the short message length will be zero.  Note
		/// that both the ShortMessage and MessagePayload cannot be used
		/// simultaneously.
		/// </summary>
		public string ShortMessage
		{
			get
			{
				return shortMessage;
			}
			
			set
			{
				shortMessage = value;
			}
		}
		
		#region optional parameters
		
		/// <summary>
		/// The message reference number assigned by the ESME.
		/// </summary>
		public UInt16 UserMessageReference
		{
			get
			{
				return GetHostOrderUInt16FromTlv((ushort)OptionalParamCodes.user_message_reference);
			}
			
			set
			{
				SetHostOrderValueIntoTlv((UInt16)Pdu.OptionalParamCodes.user_message_reference, value);
			}
		}
		
		/// <summary>
		/// The port number associated with the source address of the message.  This
		/// parameter will be present for WAP applications.
		/// </summary>
		public UInt16 SourcePort
		{
			get
			{
				return GetHostOrderUInt16FromTlv((ushort)OptionalParamCodes.source_port);
			}
			
			set
			{
				SetHostOrderValueIntoTlv((UInt16)Pdu.OptionalParamCodes.source_port, value);
			}
		}
		
		/// <summary>
		/// The port number associated with the destination address of the message.  This
		/// parameter will be present for WAP applications.
		/// </summary>
		public UInt16 DestinationPort
		{
			get
			{
				return GetHostOrderUInt16FromTlv((ushort)OptionalParamCodes.destination_port);
			}
			
			set
			{
				SetHostOrderValueIntoTlv((UInt16)Pdu.OptionalParamCodes.destination_port, value);
			}
		}
		
		/// <summary>
		/// The reference number for a particular concatenated short message.
		/// </summary>
		public UInt16 SarMsgRefNumber
		{
			get
			{
				return GetHostOrderUInt16FromTlv((ushort)OptionalParamCodes.sar_msg_ref_num);
			}
			
			set
			{
				SetHostOrderValueIntoTlv((UInt16)Pdu.OptionalParamCodes.sar_msg_ref_num, value);
			}
		}
		
		/// <summary>
		/// Total number of short message fragments within the concatenated short message.
		/// </summary>
		public byte SarTotalSegments
		{
			get
			{
				return GetOptionalParamBytes((ushort)OptionalParamCodes.sar_total_segments)[0];
			}
			
			set
			{
				SetOptionalParamBytes(
					(ushort)Pdu.OptionalParamCodes.sar_total_segments, new Byte[] {value});
			}
		}
		
		/// <summary>
		/// The sequence number of a particular short message fragment within the 
		/// concatenated short message.
		/// </summary>
		public byte SarSegmentSeqnum
		{
			get
			{
				return GetOptionalParamBytes((ushort)OptionalParamCodes.sar_segment_seqnum)[0];
			}
			
			set
			{
				SetOptionalParamBytes(
					(ushort)Pdu.OptionalParamCodes.sar_segment_seqnum, new Byte[] {value});
			}
		}
		
		/// <summary>
		/// A user response code. The actual response codes are SMS application specific.
		/// </summary>
		public byte UserResponseCode
		{
			get
			{
				return GetOptionalParamBytes((ushort)OptionalParamCodes.user_response_code)[0];
			}
			
			set
			{
				SetOptionalParamBytes(
					(ushort)Pdu.OptionalParamCodes.user_response_code, new Byte[] {value});
			}
		}
		
		/// <summary>
		/// Indicates a level of privacy associated with the message.
		/// </summary>
		public PrivacyType PrivacyIndicator
		{
			get
			{
				return(PrivacyType)
					GetOptionalParamBytes((ushort)OptionalParamCodes.privacy_indicator)[0];
			}
			
			set
			{
				SetOptionalParamBytes(
					(ushort)Pdu.OptionalParamCodes.privacy_indicator, new Byte[] {(byte)value});
			}
		}
		
		/// <summary>
		/// Defines the type of payload(e.g. WDP, WCMP, etc.)
		/// </summary>
		public PayloadTypeType PayloadType
		{
			get
			{
				return(PayloadTypeType)
					GetOptionalParamBytes((ushort)OptionalParamCodes.payload_type)[0];
			}
			
			set
			{
				SetOptionalParamBytes(
					(ushort)Pdu.OptionalParamCodes.payload_type, new Byte[] {(byte)value});
			}
		}
		
		/// <summary>
		/// This can hold up to 64K octets of short message data.
		/// The actual limit is network/SMSC dependent.
		/// </summary>
		public string MessagePayload
		{
			get
			{
				return GetOptionalParamString((ushort)OptionalParamCodes.message_payload);
			}
			
			set
			{
				PduUtil.SetMessagePayload(this, value);
			}
		}
		
		/// <summary>
		/// Associates a callback number with a message.  See section 5.3.2.36 of the
		/// SMPP spec for details.  This must be between 4 and 19 characters in length.
		/// </summary>
		public string CallbackNum
		{
			get
			{
				return GetOptionalParamString((ushort)OptionalParamCodes.callback_num);
			}
			
			set
			{
				PduUtil.SetCallbackNum(this, value);
			}
		}
		
		/// <summary>
		/// Specifies a source subaddress associated with the originating entity.
		/// See section 5.3.2.15 of the SMPP spec for details on setting this parameter.
		/// </summary>
		public string SourceSubaddress
		{
			get
			{
				return GetOptionalParamString((ushort)OptionalParamCodes.source_subaddress);
			}
			
			set
			{
				PduUtil.SetSourceSubaddress(this, value);
			}
		}
		
		/// <summary>
		/// Specifies a source subaddress associated with the receiving entity.
		/// See section 5.3.2.15 of the SMPP spec for details on setting this parameter.
		/// </summary>
		public string DestinationSubaddress
		{
			get
			{
				return GetOptionalParamString((ushort)OptionalParamCodes.dest_subaddress);
			}
			
			set
			{
				PduUtil.SetDestSubaddress(this, value);
			}
		}
		
		/// <summary>
		/// The language of the short message.
		/// </summary>
		public LanguageType LanguageIndicator
		{
			get
			{
				return(LanguageType)
					GetOptionalParamBytes((ushort)OptionalParamCodes.language_indicator)[0];
			}
			
			set
			{
				SetOptionalParamBytes(
					(ushort)Pdu.OptionalParamCodes.language_indicator, new Byte[] {(byte)value});
			}
		}
		
		/// <summary>
		/// From the SMPP spec:
		/// The its_session_info parameter is a required parameter for the CDMA Interactive
		/// Teleservice as defined by the Korean PCS carriers [KORITS]. It contains control
		/// information for the interactive session between an MS and an ESME.
		///
		/// See section 5.3.2.43 of the SMPP spec for how to set this.
		/// </summary>
		public string ItsSessionInfo
		{
			get
			{
				return GetOptionalParamString((ushort)OptionalParamCodes.its_session_info);
			}
			
			set
			{
				PduUtil.SetItsSessionInfo(this, value);
			}
		}
		
		/// <summary>
		/// Network Error Code.  May be present for Intermediate Notifications
		/// and SMSC Delivery Receipts.  See SMPP spec 5.3.2.31 for details.
		/// </summary>
		public string NetworkErrorCode
		{
			get
			{
				return GetOptionalParamString((ushort)OptionalParamCodes.network_error_code);
			}
			
			set
			{
				PduUtil.SetNetworkErrorCode(this, value);
			}
		}
		
		/// <summary>
		/// Indicates to the ESME the final message state for an SMSC Delivery Receipt.
		/// </summary>
		public MessageStateType MessageState
		{
			get
			{
				return(MessageStateType)
					GetOptionalParamBytes((ushort)OptionalParamCodes.message_state)[0];
			}
			
			set
			{
				SetOptionalParamBytes(
					(ushort)Pdu.OptionalParamCodes.message_state, new Byte[] {(byte)value});
			}
		}
		
		/// <summary>
		/// Indicates the ID of the message being receipted in an SMSC Delivery Receipt.
		/// </summary>
		public string ReceiptedMessageId
		{
			get
			{
				return GetOptionalParamString((ushort)OptionalParamCodes.receipted_message_id);
			}
			
			set
			{
				PduUtil.SetReceiptedMessageId(this, value);
			}
		}
		
		#endregion optional parameters
		
		#region constructors
		
		/// <summary>
		/// Creates a deliver_sm Pdu.
		/// </summary>
		/// <param name="incomingBytes">The bytes received from an ESME.</param>
		public SmppDeliverSm(byte[] incomingBytes): base(incomingBytes)
		{}
		
		/// <summary>
		/// Creates a deliver_sm Pdu.
		/// </summary>
		public SmppDeliverSm(): base()
		{}
		
		#endregion constructors
		
		/// <summary>
		/// This decodes the deliver_sm Pdu.  The Pdu has basically the same format as
		/// the submit_sm Pdu, but in this case it is a response.
		/// </summary>
		protected override void DecodeSmscResponse()
		{
			byte[] remainder = BytesAfterHeader;
			ServiceType = SmppStringUtil.GetCStringFromBody(ref remainder);
			SourceAddressTon = (TonType)remainder[0];
			SourceAddressNpi = (NpiType)remainder[1];
			SourceAddress = SmppStringUtil.GetCStringFromBody(ref remainder, 2);
			DestinationAddressTon = (TonType)remainder[0];
			DestinationAddressNpi = (NpiType)remainder[1];
			DestinationAddress = SmppStringUtil.GetCStringFromBody(ref remainder, 2);
			EsmClass = remainder[0];
			ProtocolId = (SmppVersionType)remainder[1];
			PriorityFlag = (PriorityType)remainder[2];
			//schedule_delivery_time and validity_period are null, so don't bother
			//reading them
			RegisteredDelivery = (RegisteredDeliveryType)remainder[5];
			//replace_if_present is always null, so don't bother reading it
			DataCoding = (DataCodingType)remainder[7];
			//sm_default_msg_id is always null, so don't bother reading it
			smLength = remainder[9];
			ShortMessage = SmppStringUtil.GetStringFromBody(ref remainder, 10, 10 + smLength);
			TranslateTlvDataIntoTable(remainder);
		}
		
		/// <summary>
		/// Initializes this Pdu.
		/// </summary>
		protected override void InitPdu()
		{
			base.InitPdu();
			CommandStatus = 0;
			CommandID = CommandIdType.deliver_sm;
		}
		
		///<summary>
		/// Gets the hex encoding(big-endian)of this Pdu.
		///</summary>
		///<return>The hex-encoded version of the Pdu</return>
		public override void ToMsbHexEncoding()
		{
			ArrayList pdu = GetPduHeader();
			pdu.AddRange(SmppStringUtil.ArrayCopyWithNull(Encoding.ASCII.GetBytes(ServiceType)));
			pdu.Add((byte)SourceAddressTon);
			pdu.Add((byte)SourceAddressNpi);
			pdu.AddRange(SmppStringUtil.ArrayCopyWithNull(Encoding.ASCII.GetBytes(SourceAddress)));
			pdu.Add((byte)DestinationAddressTon);
			pdu.Add((byte)DestinationAddressNpi);
			pdu.AddRange(SmppStringUtil.ArrayCopyWithNull(Encoding.ASCII.GetBytes(DestinationAddress)));
			pdu.Add(EsmClass);
			pdu.Add((byte)ProtocolId);
			pdu.Add((byte)PriorityFlag);
			//schedule_delivery_time and validity_period are null, so set them to zero
			pdu.Add((byte)0);
			pdu.Add((byte)0);
			pdu.Add((byte)RegisteredDelivery);
			//replace_if_present is always null, so set it to zero
			pdu.Add((byte)0);
			pdu.Add((byte)DataCoding);
			//sm_default_msg_id is always null, so set it to zero
			pdu.Add((byte)0);
			smLength = PduUtil.InsertShortMessage(pdu, ShortMessage);
			
			PacketBytes = EncodePduForTrasmission(pdu);
		}
	}
}
