using System;
using System.Collections;
using System.Net;
using System.Text;
using System.Diagnostics;

namespace SMPP.Utility
{
	/// <summary>
	/// Tag, length, value table for SMPP PDUs.  The methods in this class assume 
	/// that the tag passed in is already in network byte order.
	/// </summary>
	public class TlvTable
	{
		private Hashtable tlvTable;
		
		
		public TlvTable()
		{
			tlvTable = new Hashtable();
			tlvTable = Hashtable.Synchronized(tlvTable);
		}
		
		
		public void TranslateTlvDataIntoTable(byte[] tlvData)
		{
			TranslateTlvDataIntoTable(tlvData, 0);
		}
		
		
		public void TranslateTlvDataIntoTable(byte[] tlvData, Int32 index)
		{
			if(tlvData == null || tlvData.Length <= 0)
			{
				return;
			}
			
			//go through and decode the TLVs
			while(index < tlvData.Length)
			{
				InjectTlv(tlvData, ref index);
			}
		}
		
		
		private void InjectTlv(byte[] tlvData, ref Int32 index)
		{
			byte[] temp = new byte[2];
			temp[0] = tlvData[index];
			temp[1] = tlvData[index + 1];
			
			#if DEBUG
			Console.WriteLine("tag bytes " + temp[0].ToString("X").PadLeft(2, '0') + 
			                  temp[1].ToString("X").PadLeft(2, '0'));
			#endif
			
			UInt16 tag = BitConverter.ToUInt16(temp, 0);
			index += 2;
			temp[0] = tlvData[index];
			temp[1] = tlvData[index + 1];
			UInt16 length = UnsignedNumConverter.SwapByteOrdering(BitConverter.ToUInt16(temp, 0));
			index += 2;
			//decode the value
			
			#if DEBUG
			Console.WriteLine("TLV Length " + length);
			#endif
			
			ArrayList data = new ArrayList(length);
			
			int total = index + length;
			for(int k = index;(k < index + length)&& k < tlvData.Length; k++)
			{
				data.Add(tlvData[k]);
			}
			
			data.TrimToSize();
			//add the values to the hashtable
			tlvTable.Add(tag, data.ToArray(typeof(byte)));
			//set it up for the next run
			index += length;
		}
		
		
		public byte[] GetOptionalParamBytes(UInt16 tag)
		{
			object val = tlvTable[tag];
			if(val == null)
			{
				throw new ApplicationException("TLV tag " + tag + " not found.");
			}
			else
			{
				byte[] bVal = (byte[])val;
				#if DEBUG
				StringBuilder sb = new StringBuilder();
				sb.Append("Getting tag " + UnsignedNumConverter.SwapByteOrdering(tag));
				sb.Append("\nValue: ");
				
				for(int i = 0; i < bVal.Length; i++) 
				{
					sb.Append(bVal[i]);
					sb.Append(" ");
				}
				
				Console.WriteLine(sb);
				
				#endif
				return bVal;
			}
		}
		
		
		public string GetOptionalParamString(UInt16 tag)
		{
			byte[] val = GetOptionalParamBytes(tag);
			
			return Encoding.ASCII.GetString(val);
		}
		
		
		public void SetOptionalParamBytes(UInt16 tag, byte[] val)
		{
			#if DEBUG
			StringBuilder sb = new StringBuilder();
			sb.Append("Setting tag " + UnsignedNumConverter.SwapByteOrdering(tag));
			sb.Append("\nValue: ");
			
			for(int i = 0; i < val.Length; i++) 
			{
				sb.Append(val[i]);
				sb.Append(" ");
			}
			
			Console.WriteLine(sb);
			
			#endif
			if(val != null)
			{
				if(val.Length > UInt16.MaxValue)
				{
					throw new Exception("Optional parameter value for " + tag + " is too large.");
				}
				tlvTable.Add(tag, val);
			}
		}
		
		
		public void SetOptionalParamString(UInt16 tag, string val)
		{
			if(val != null)
			{
				SetOptionalParamBytes(tag, Encoding.ASCII.GetBytes(val));
			}
		}
		
		
		public void UpdateOptionalParamBytes(UInt16 tag, byte[] val)
		{
			object obj = tlvTable[tag];
			
			if(obj != null)
			{
				tlvTable[tag] = val;
			}
			else
			{
				throw new ApplicationException("TLV tag " + tag + " not found");
			}
		}
		
		
		public void UpdateOptionalParamString(UInt16 tag, string val)
		{
			UpdateOptionalParamBytes(tag, Encoding.ASCII.GetBytes(val));
		}
		
		
		public ArrayList GenerateByteEncodedTlv()
		{
			if(tlvTable == null || tlvTable.Count <= 0)
			{
				return new ArrayList(0);
			}
			
			ArrayList tlvs = new ArrayList();
			IDictionaryEnumerator iterator = tlvTable.GetEnumerator();
			ArrayList elem = new ArrayList();
			while(iterator.MoveNext())
			{
				elem.Clear();
				//tag-2 bytes
				elem.AddRange(BitConverter.GetBytes(((UInt16)iterator.Key)));
				//length-2 bytes
				byte[] nextVal = (byte[])iterator.Value;
				UInt16 tlvLength = UnsignedNumConverter.SwapByteOrdering(((UInt16)(nextVal).Length));
				elem.AddRange(BitConverter.GetBytes(tlvLength));
				//value
				elem.AddRange(nextVal);
				elem.TrimToSize();
				//copy it over to a byte array
				tlvs.Add(elem.ToArray(typeof(byte)));
			}
			tlvs.TrimToSize();
			
			return tlvs;
		}
	}
}
